/*
KabooM - A rewrite of the game KabooM! by Activision for the iPod.
Copyright (C) 2005 Daniel Smith

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

General Game Information:

Prepare yourself for a supreme test of reflexes, coordination, and
agility.  You're about to face the world's most unpredictable and
relentless "Mad Bomber."  He hates losing as much as you love winning.
So, to keep him frowning, take a minute to read over these instructions.
Then, grab your buckets and bomb away!

By turning your IPod Wheel clockwise, you move your
buckets of water to the right.  Turning the Wheel counterclockwise
moves buckets to the left.

Remeber: The objective of the game is to catch all of the bombs, as well as
having fun!.
*/

#include "pz.h"
#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>

static PzModule *module;
static ttk_surface man_img, bucket_img, bomb_img;

typedef struct {
    int x;
    int y;
    int width;
    int height;
} rectangle;

typedef struct {
    int bomb_speed;
    int clust;
    int bomb_amount;
} info;

static rectangle bucket[3];

static int bombdrop;		/*the timer thingo for how long to drop the bomb(s), which is defined in new_kaboom window... */

static int man_direction;	/*direction for man... */
static int direction_time;	/*the time to go for direction */
static int man_location_x;	/*location x for man */
static int man_location_y;
static int man_width;
static int bomb_size;
static int man_height;
static int bomb_caught;		/*coutner for how many caught bombs */
static int lives;		/* counter for how many lives which are left */
static int buckets_alive;	/* counter for how many buckets are alive */
static info level_info[9];	/*info about the the current level */
static int level;
static int bombs_counted;	/*counter for how many bombs have been counted in that level, so that                                       the level rotation works according to plan */


struct bomblist {
    int x;
    int y;
    struct bomblist *next;
};

struct bomblist *bomblist_head;

static void new_game();

static void bomber_timeout()
{
    bombs_counted = 0;
    direction_time = 5000;
    man_direction = 0;
    bombdrop = 5000;
}

static void free_bomblist()
{
    struct bomblist *cur, *next;
    cur = bomblist_head;
    while (cur) {
	next = cur->next;
	free(cur);
	cur = next;
    }
    bomblist_head = 0;
}

static void bomb_draw (ttk_surface srf)
{				/*function to draw the bomb */
    struct bomblist *cur = bomblist_head;
    struct bomblist *last = 0;
    while (cur) {
	int bucket_hits[3], bucket_hit = 0, offscreen = 0, i;

        ttk_blit_image (bomb_img, srf, cur->x, cur->y);

	for (i = 0; i < 3; i++) {
	    bucket_hits[i] = (lives &&
			      (bucket[i].x - (bucket[i].width / 2) <=
			       cur->x)
			      &&
			      ((bucket[i].x - (bucket[i].width / 2) +
				(bucket[i].width * 2)) >=
			       (cur->x + bomb_size))
			      && (bucket[i].y <= (cur->y + bomb_size))
			      && ((cur->y + bomb_size) <=
				  (bucket[i].y + bucket[i].width))
		);
	}

	if (cur->y >= (ttk_screen->h - ttk_screen->wy))
	    offscreen = 1;
	bucket_hit = bucket_hits[0] || bucket_hits[1] || bucket_hits[2];

	if (bucket_hit) {
	    bomb_caught += level_info[level].bomb_amount;
	    bombs_counted++;

	} else if (offscreen) {
	    bomber_timeout();
	    buckets_alive--;
	    bombs_counted = 0;
	    lives--;

	    if (!buckets_alive) {
		pz_message ("GAME OVER!");
                new_game();
                bomber_timeout();
                break;
	    }
	}


	if (bucket_hit || offscreen) {
	    if (!last) {
		bomblist_head = cur->next;
	    } else {
		last->next = cur->next;
	    }
	    free(cur);
	    if (last)
		cur = last->next;
	    else
		cur = bomblist_head;
	} else {
	    cur->y += level_info[level].bomb_amount;
	    last = cur;
	    cur = cur->next;
	}
    }

}



static void bomb_add()
{
    struct bomblist *newBomb = malloc(sizeof(struct bomblist));

    newBomb->x = man_location_x;
    newBomb->y = (ttk_screen->wy * 4 / 3);
    newBomb->next = bomblist_head;
    bomblist_head = newBomb;
};



void wall_draw (ttk_surface srf)
{				/*function to draw the wall which the robber guy sits on, which includes the colours */

    ttk_line (srf, 0, ttk_screen->wy * 4 / 3, ttk_screen->w, ttk_screen->wy * 4 / 3, ttk_makecol (BLACK));

    if (ttk_screen->bpp == 16) {
        ttk_fillrect (srf, 0, ttk_screen->wy * 4 / 3, ttk_screen->w, ttk_screen->h,
                      ttk_makecol (0, 100, 0));
    }
}


void bucket_draw (ttk_surface srf)
{				/*functioon to draw the bucket */
    switch (buckets_alive) {
        /* FALLTHRUs */
        case 3: ttk_blit_image (bucket_img, srf, bucket[0].x, bucket[0].y);
        case 2: ttk_blit_image (bucket_img, srf, bucket[1].x, bucket[1].y);
        case 1: ttk_blit_image (bucket_img, srf, bucket[2].x, bucket[2].y);
        default: break;
    }
}

static void man_draw (ttk_surface srf)
{				/*function to draw the man who drops the bombs */
    man_location_x += man_direction;
    if (man_location_x <= 0 && man_direction < 0)
	man_direction *= -1;
    if (man_location_x >= ttk_screen->w - man_width && man_direction > 0) {
	man_location_x = ttk_screen->w - man_width;
	man_direction *= -1;
    }

    ttk_blit_image (man_img, srf, man_location_x, man_location_y);
}


static void kaboom_draw (TWidget *this, ttk_surface srf)
{
    char text[30];
    wall_draw (srf);		/* calls the drawing of the wall (which is really a line for now) */
    bomb_draw (srf);		/* call to draw the bomb */
    bucket_draw (srf);		/*call to draw the bucket */
    man_draw (srf);			/*call to draw the man who drops bombs */

    /* Draw the counter */
    sprintf(text, "Score: %d", bomb_caught);
    ttk_text (srf, ttk_textfont, 1, ttk_screen->wy - ttk_text_height (ttk_textfont), ttk_makecol (BLACK), text);
}

static void kaboom_loop (TWidget *this)
{
    bombdrop -= 50;
    if (bombdrop == 0) {
        bomb_add();
        bombdrop = level_info[level].bomb_speed;
    }
    
    direction_time -= 50;
    if (direction_time <= 0) {
        direction_time = rand() % 800 + 200;
        man_direction = rand() % 11 - 5;
    }

    if (bombs_counted >= level_info[level].clust) {
	level++;
	bomber_timeout();
    }

    this->dirty = 1;
}

static int kaboom_event(PzEvent * ev)
{
    int ret = 0;
    static int paused = 0;

    switch (ev->type) {
    case PZ_EVENT_TIMER:
	if (!paused)
	    kaboom_loop (ev->wid);
	break;

    case PZ_EVENT_SCROLL:
        bucket[0].x += 5*ev->arg;
        bucket[1].x += 5*ev->arg;
        bucket[2].x += 5*ev->arg;
        if (bucket[2].x < 0) bucket[0].x = bucket[1].x = bucket[2].x = 0;
        if (bucket[2].x >= (ev->wid->w - bucket[2].width)) {
            bucket[0].x = bucket[1].x = bucket[2].x = ev->wid->w - bucket[2].width;
        }
        ev->wid->dirty = 1;
        break;
        
    case PZ_EVENT_BUTTON_DOWN:
        switch (ev->arg) {
	case PZ_BUTTON_HOLD:	/* hold switch on */
	    paused = 1;		/* pause the loop */
            ev->wid->dirty = 1;
	    break;
	case PZ_BUTTON_MENU:	/* Menu button */
	    pz_close_window (ev->wid->win);
	    break;
	default:
	    ret |= TTK_EV_UNUSED;	/* allow a default key action */
	    break;
	}
	break;
    case PZ_EVENT_BUTTON_UP:
	switch (ev->arg) {
	case PZ_BUTTON_HOLD:	/* un-held */
	    paused = 0;
            ev->wid->dirty = 1;
	    break;
	default:
	    ret |= TTK_EV_UNUSED;
	    break;
	}
    default:
	ret |= TTK_EV_UNUSED;	/* we didn't ask for this event, go away */
    }
    return ret;
}


static void new_game() 
{
    bucket[0].x = bucket[1].x = bucket[2].x = (ttk_screen->wy / 2);	/* location x for buckets */
    bucket[0].y = (((ttk_screen->h - ttk_screen->wy) * 95) / 100);	/* location y for bucket0 */
    bucket[1].y = (((ttk_screen->h - ttk_screen->wy) * 85) / 100);	/* location y for bucket1 */
    bucket[2].y = (((ttk_screen->h - ttk_screen->wy) * 75) / 100);	/* location y for bucket0 */

    man_location_x = (ttk_screen->w / 2);	/*location x for man */
    direction_time = rand() % 300;
    man_direction = rand() % 11 - 5;
    bomb_caught = 0;		/*counter for how many caught bombs */
    bombdrop = 1000;
    buckets_alive = 3;
    lives = 3;
    level = 1;
    bombs_counted = 0;

    free_bomblist();
}

static TWindow *new_kaboom_window(void)
{
    new_game();

    return pz_do_window ("KabooM!", PZ_WINDOW_NORMAL, kaboom_draw, kaboom_event, 50);
}

void init_kaboom() 
{
    ttk_surface tmp;
    int w, h;

    /* PZ inits */
    module = pz_register_module ("kaboom", 0);
    pz_menu_add_action ("/Extras/Games/Kaboom", new_kaboom_window);

    /* Kaboom inits, for stuff that doesn't change */
    bucket[0].width = bucket[1].width = bucket[2].width = ttk_screen->w / 10;	/*bucket width */
    bucket[0].height = bucket[1].height = bucket[2].height = ttk_screen->w / 30;	/*bucket height */
    man_width = ttk_screen->w / 16;
    man_height = ttk_screen->wy * 4 / 3;
    bomb_size = (man_width * 3 / 4);
    man_location_y = 1;

    /** Level info */
    /* The next part defines the number of bombs per cluster, without
     * ten lines of stars.
     */
    level_info[0].clust = 0;
    level_info[1].clust = 10;
    level_info[2].clust = 20;
    level_info[3].clust = 30;
    level_info[4].clust = 40;
    level_info[5].clust = 50;
    level_info[6].clust = 75;
    level_info[7].clust = 100;
    level_info[8].clust = 150;

    /* All of this stuff below defines the speed at which the man releases
     * bombs, without unnecessary capitalization.
     */
    level_info[0].bomb_speed = 0;
    level_info[1].bomb_speed = 1000;
    level_info[2].bomb_speed = 900;
    level_info[3].bomb_speed = 800;
    level_info[4].bomb_speed = 700;
    level_info[5].bomb_speed = 600;
    level_info[6].bomb_speed = 500;
    level_info[7].bomb_speed = 400;
    level_info[8].bomb_speed = 300;

    /* All of this stuff below *defines* the speed at which the bombs fall
     * (not DEFINES, and not definese).
     */
    level_info[0].bomb_amount = 0;
    level_info[1].bomb_amount = 1;
    level_info[2].bomb_amount = 2;
    level_info[3].bomb_amount = 3;
    level_info[4].bomb_amount = 4;
    level_info[5].bomb_amount = 5;
    level_info[6].bomb_amount = 6;
    level_info[7].bomb_amount = 7;
    level_info[8].bomb_amount = 8;

    /* Image loading and scaling */
    tmp = ttk_load_image (pz_module_get_datapath (module, "man.png"));
    ttk_surface_get_dimen (tmp, &w, &h);
    man_img = ttk_scale_surface (tmp, (float)man_width / (float)w);
    ttk_free_surface (tmp);

    tmp = ttk_load_image (pz_module_get_datapath (module, "bomb.png"));
    ttk_surface_get_dimen (tmp, &w, &h);
    bomb_img = ttk_scale_surface (tmp, (float)bomb_size / (float)w);
    ttk_free_surface (tmp);

    tmp = ttk_load_image (pz_module_get_datapath (module, "bucket.png"));
    ttk_surface_get_dimen (tmp, &w, &h);
    bucket_img = ttk_scale_surface (tmp, (float)bucket[0].width / (float)w);
    ttk_free_surface (tmp);
}

PZ_MOD_INIT (init_kaboom)
